
/*********************************************************************
* Copyright (c) 2018 Tecnalia.
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
* Contributors (in alphabetical order):
* Alberto Molinuevo                         Tecnalia
* Gorka Benguria                            Tecnalia
* Iaki Etxaniz                             Tecnalia
* Juncal Alonso                             Tecnalia
* Leire Orue-Echevarria                     Tecnalia
* Maria Jose Lopez                          Tecnalia
* Marisa Escalante                          Tecnalia
* Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
**********************************************************************/
package eu.DECIDEh2020.optimus.editors;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Paths;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
//import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.wb.swt.SWTResourceManager;
//import org.eclipse.swt.widgets.Table;
import org.json.simple.JSONArray;

import eu.DECIDEh2020.appManager.AppManager;
import eu.DECIDEh2020.appManager.HistoryManager;
import eu.DECIDEh2020.appManager.exceptions.AppManagerException;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;



//import com.fasterxml.jackson.databind.JsonNode;

import eu.DECIDEh2020.appManager.models.AppDescription;
import eu.DECIDEh2020.appManager.models.HistoricalDeployment;
import eu.DECIDEh2020.appManager.models.HistoryEntry;
import eu.DECIDEh2020.appManager.models.SchemaElement;
import eu.DECIDEh2020.architect.plugin.editor.Page;
//import eu.decideh2020.ACSmI.client.model.Service;
import eu.decideh2020.optimus.client.api.DefaultApi;
import eu.decideh2020.optimus.client.model.ReturnedSimulation;
import eu.decideh2020.optimus.client.model.ReturnedSc;
import eu.decideh2020.optimus.client.model.ReturnedSchema;
import eu.decideh2020.optimus.client.model.Simulation;
import com.fasterxml.jackson.jaxrs.json.JacksonJsonProvider;

//import javax.ws.rs.core.Response;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.ClientConfiguration;
import org.apache.cxf.jaxrs.client.WebClient;
import org.eclipse.swt.widgets.Text;



public class SimulateSchema extends Page {



	/**
	 * Create the composite.
	 * @param parent
	 * @param style
	 */
	private DefaultApi apiInstance;
	private JSONArray NFRs;
	private AppDescription appdescription;
	private Text simText;
	
	public SimulateSchema(Composite parent, int style, AppDescription appdescription) {
		super(parent, SWT.NONE, appdescription);
		this.appdescription = appdescription;
		createSimulationGUI(appdescription);
		

	}
	
	
	private void createSimulationGUI(AppDescription appd) {
		// TODO Auto-generated method stub
		Control[] children = this.getChildren();
	    for (int i = 0 ; i < children.length; i++) {
	        children[i].dispose();
	    }

		ScrolledComposite scrolledComposite = new ScrolledComposite(this, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
		scrolledComposite.setAlwaysShowScrollBars(true);
		scrolledComposite.setBounds(10, 43, 982, 680);
		scrolledComposite.setExpandHorizontal(true);
		scrolledComposite.setExpandVertical(true);
		
		Group group = new Group(scrolledComposite, SWT.NONE);
		NFRs = new JSONArray();
		
		
		Button btnSimulate = new Button(group, SWT.NONE);
		btnSimulate.setBounds(10, 10, 75, 25);
		btnSimulate.setText("Simulate");
		
		Composite composite_1 = new Composite(group, SWT.NONE);
		composite_1.setBounds(0, 51, 901, 596);
		
		Label lblNewLabel = new Label(composite_1, SWT.NONE);
		lblNewLabel.setFont(SWTResourceManager.getFont("Calibri", 14, SWT.BOLD));
		lblNewLabel.setBounds(10, 10, 484, 28);
		lblNewLabel.setText("Best Deployment Schema obtained by Simulation:");
		
		simText = new Text(composite_1, SWT.BORDER | SWT.WRAP | SWT.MULTI);
		simText.setBounds(10, 62, 1070, 507);

		
		this.apiInstance = apiInstance;
        simText.setText("No simulation launched");
		btnSimulate.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent event) {
			

			Simulation simL = new Simulation();
			simL.setAppurl("Appurl");
			

			JacksonJsonProvider provider = new JacksonJsonProvider();
			
	        List providers = new ArrayList();
	        providers.add(provider);
	        
	        
	        apiInstance = JAXRSClientFactory.create("URLoptimusserver", DefaultApi.class, providers);
	        
	        
	        org.apache.cxf.jaxrs.client.Client client = WebClient.client(apiInstance);
	        
	        ClientConfiguration config = WebClient.getConfig(client); 
			//try {
	        
	        List<ReturnedSimulation> response = new ArrayList<ReturnedSimulation>();
	        response = apiInstance.createsSimulationUsingPOST(simL);
	        //ReturnedSimulation simToShow = new ReturnedSimulation();
	        Integer id = response.get(0).getSimid();
	       
	        //simulationText.setText("This Simulation has been launched: \n"+id);
	        simText.setText("This Simulation has been launched: \n"+id);
	        
	        
	        
	        String status = response.get(0).getStatus();
	        while (status.equals("started")) {
	        	//response = apiInstance.getSimulationUsingGET(id);
	        	List<ReturnedSimulation> response2 = new ArrayList<ReturnedSimulation>();
	        	
	        	response2 = apiInstance.getSimulationUsingGET(id);
	        	status = response2.get(0).getStatus();
	        }
	        List<ReturnedSc> response2 = new ArrayList<ReturnedSc>();
        	response2 = apiInstance.getSchemaUsingGET(id);
	        List<SchemaElement> schemas = new ArrayList<>();
	        //at the moment i consider the best of all of them, the first one.
	        //when integration--> increase index from 0 to 5.
			for (ReturnedSc rs : response2) {
				SchemaElement sc = new SchemaElement();
				sc.setCsId(rs.getCsId());
				sc.setIndex(rs.getIndex());
				sc.setMicroservices(rs.getMicros());
				schemas.add(sc);
			}
			
			//store the schema (5 schemas later) into the history repo
			try {
				storeHistorySchemas(schemas,id);
			} catch (AppManagerException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
	       
	       		appdescription.setSchema(schemas);
		        //List<SchemaElement> schemas2 = new ArrayList<>();
		        //schemas2 = appdescription.getSchema();
		        String scText = "";
		        if (!schemas.isEmpty())
		        {
		        scText = "The best Schema found for the simulation is:\n\n";
		        for (SchemaElement se : schemas) {
		    		scText = scText+"    The following microservices: \n"+se.getMicroservices().toString()+"\n\n    will be deployed in the following Cloud Service: "+se.getCsId()+"\n\n";
		    	}
		        //simulationText.setText(scText);
		        simText.setText(scText);
		        }


			//simulationText.setText("Simulation launched from btnSimulate.....");
			addDisposeListener(new DisposeListener() {
				public void widgetDisposed(DisposeEvent e) {
				}
			});
			
		}

		private void storeHistorySchemas(List<SchemaElement> schemas, Integer id) throws AppManagerException, IOException {
			AppManager manager = null;
			String localAppDesc = System.getProperty("java.io.tmpdir");
			localAppDesc = localAppDesc+"\\"+id;
			//localAppDesc = localAppDesc+"\\"+System.currentTimeMillis();
			//this is the way to manage the variables (paths only) by java, obtaining tmp system folder - end
			
			//System.out.println("my applocalfolder is: "+Paths.get(localAppDesc));
			try {
				manager = AppManager.open(Paths.get(localAppDesc));

			} catch (AppManagerException e) {
				e.printStackTrace();
			}
			HistoryManager historyManager = manager.getHistoryManager();
			List<HistoryEntry> history = new ArrayList();
			history = historyManager.getHistory();

			HistoryEntry he = new HistoryEntry();
			List<SchemaElement> deployments = new ArrayList();
			he.setDate(LocalDateTime.now().toString());
			for (SchemaElement se : schemas) {
				deployments.add(se);
			}
			
			he.setSchema(deployments);
			//history.add(he);
			//for now i write it but first i should ask the developer
			//historyManager.write(history);
			
		}
	});
		scrolledComposite.setContent(group);
		scrolledComposite.setMinSize(group.computeSize(SWT.DEFAULT, SWT.DEFAULT));
		
	}
	@Override
	public void open(AppDescription appd) {
		// TODO Auto-generated method stub
		this.appdescription = appd;
		createSimulationGUI(appd);
		
	}

	   private Properties loadProperties() {
	        Properties props = new Properties();
	        
	        InputStream fis = null;
	        try {
	            fis = this.getClass().getClassLoader().getResourceAsStream("/optimus.server.properties");
	            props.load(fis);
	        }     catch (IOException e) {
	            // TODO Auto-generated catch block
	            e.printStackTrace();
	        }
	        
	        return props;
	  }
}
