/*********************************************************************
* Copyright (c) 2018 Tecnalia.
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
* Contributors (in alphabetical order):
* Alberto Molinuevo                         Tecnalia
* Gorka Benguria                            Tecnalia
* Iñaki Etxaniz                             Tecnalia
* Juncal Alonso                             Tecnalia
* Leire Orue-Echevarria                     Tecnalia
* Maria Jose Lopez                          Tecnalia
* Marisa Escalante                          Tecnalia
* Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
**********************************************************************/
package eu.decideh2020.optimus.server.api;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.UUID;

import org.apache.cxf.jaxrs.client.ClientConfiguration;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.jaxrs.json.JacksonJsonProvider;


import eu.DECIDEh2020.appManager.AppManager;
import eu.DECIDEh2020.appManager.exceptions.AppManagerException;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.AppDescription;
import eu.DECIDEh2020.appManager.models.AvailabilityNfr;
import eu.DECIDEh2020.appManager.models.DetachableResource;
import eu.DECIDEh2020.appManager.models.Microservice;
import eu.DECIDEh2020.appManager.models.Nfr;
import eu.DECIDEh2020.appManager.models.PerformanceNfr;
import eu.DECIDEh2020.appManager.models.SchemaElement;
import eu.decideh2020.optimus.server.api.impl.Attribute;
import eu.decideh2020.optimus.server.api.impl.Filter;
import eu.decideh2020.optimus.server.api.impl.SimQueue;
import eu.decideh2020.optimus.server.model.sortingPossibilities;
import io.swagger.util.Json;
import eu.decideh2020.optimus.server.model.ReturnedSc;
import eu.decideh2020.optimus.server.model.ReturnedSchema;
import eu.decideh2020.optimus.server.model.Simulation;
import eu.decideh2020.optimus.server.model.msANDcs;
import eu.decideh2020.optimus.server.model.sortingPossibilities;
import eu.decideh2020.acsmi.client.api.ServiceResourceApi;
import eu.decideh2020.acsmi.client.model.Service;

//import eu.decideh2020.optimus.server.api.JacksonJsonProvider;
import org.json.simple.JSONArray;
//import org.json.simple.JSONException;
import org.json.simple.JSONObject;




public class SimulatorThread extends Thread {

	private Simulation headofq;
	private HashMap<Integer, String> classid = new HashMap<Integer, String>();
	private ServiceResourceApi api;
	private List<msANDcs> schemasPossibilities = new ArrayList<>();
	

    List<SchemaElement> sc = new ArrayList<>();
    String localAppDesc = "";
	
	// Constructor, getter & setter

	@Override
	public void run() {
		InitializeHashmap ();
		SimQueue SimQ = SimQueue.getInstance( );
		while (true) {
			headofq = SimQ.poll();
			if (headofq!=null && headofq.getStatus().contentEquals("started")) {
				this.simprocess(headofq);
			}
			this.wait(2);
		}
	}

		

	private void InitializeHashmap() {
		classid.put(1, "db");
		classid.put(2, "storage");
		classid.put(3, "Computing");
		classid.put(3, "Computing public IP");

		
		
	}



	private void simprocess(Simulation hq) {
		//the simulation process for invoking the discovery and obtaining the schema
		//System.out.println("simulation launched for: "+hq.getSimid());
		List<Microservice> microservices = new ArrayList<>();
		List<Nfr> nfrs = new ArrayList<>();
		AppManager manager = null;
		if (hq.getAppurl().contains("https://"))
		{
			localAppDesc = System.getProperty("java.io.tmpdir");
			localAppDesc = localAppDesc+"\\"+hq.getSimid();
			//this is the way to manage the variables (paths only) by java, obtaining tmp system folder - end
			try {
				manager = AppManager.open(hq.getAppurl(), "user", "password", Paths.get(localAppDesc));

			} catch (AppManagerException e) {
				e.printStackTrace();
			}
		}
		else {
			localAppDesc = hq.getAppurl();
			try {
				manager = AppManager.open(Paths.get(localAppDesc));
			} catch (AppManagerException e) {
				e.printStackTrace();
			}
		}
		

		AppDescription description = new AppDescription();
		sc.clear();
		try {

			description = manager.getAppDescription();
			nfrs = description.getNfrs();
			microservices = description.getMicroservices();
			description.getSchema().clear();
			for (Microservice ms : microservices) {

				BuildRequest(ms,nfrs, description);
			}
			
		} catch (IOException | DECIDEValidationException e) {
			e.printStackTrace();
		}
		description.getSchema().clear();
		BestSchema(hq, description, manager);
		
	}



	private void BestSchema(Simulation sim, AppDescription appd, AppManager manager) {
        //obtain the best possible schema taking into account the information sent by ACSmI
		// set the status to the simulation to finished
		//sorting the individual schemas possibilities
		sortingPossibilities sorter = new sortingPossibilities();
	        sorter.sort(schemasPossibilities);
	        //combine the individual schemas possibilities
	        sc = sorter.combine(schemasPossibilities);
	        sim.setStatus("finished");
	        if (!localAppDesc.contains("https://"))
		{
	        String scText = "";
	        List<ReturnedSc> schemas = new ArrayList<>();
	        if (!sc.isEmpty())
	        {
	        
	        for (SchemaElement se : sc) {
	    		ReturnedSc rs = new ReturnedSc();
	    		rs.setIndex(se.getIndex());
	    		rs.setCsId(se.getCsId());
	    		for (String micro : se.getMicroservices()) {
	    		rs.addMicrosItem(micro);}
	    		schemas.add(rs);
	    		
	    	}

	        sim.setSc(schemas);
			SimQueue SimQ = SimQueue.getInstance( );
			SimQ.add(sim);
			if (sim.getCaller()=="VH") {
				//record the appdescription into de git
			}
	        }
		}
	}
	//}
	private void BuildRequest(Microservice ms, List<Nfr> nfrs, AppDescription appdesc) {
		Filter filterexample = new Filter();
		List<Attribute> attributes = new ArrayList<>();
		// all the CS are VM
		filterexample.setClassid(3);
		for (Nfr nf : nfrs) {
			
			if (nf instanceof PerformanceNfr) {
				PerformanceNfr performanceNfr = (PerformanceNfr) nf;
				Attribute att = new Attribute();
				att.setTypeid(21);
				att.setName("Response Time: Virtual Machine Performance");
				if (performanceNfr.getUnit().contains("percentage")) {
					att.setUnit("%");
				}
				if (performanceNfr.getUnit().contains("milliseconds")) {
					att.setUnit("ms");
				}
				att.setValue(Double.toString(performanceNfr.getValue()));
				attributes.add(att);
			}
               if (nf instanceof AvailabilityNfr) {
                      AvailabilityNfr availabilityNfr = (AvailabilityNfr) nf;
                      
				Attribute att = new Attribute();
				att.setTypeid(20);
				att.setName("Availability");
				att.setValue(Double.toString(availabilityNfr.getValue()));
				if (availabilityNfr.getUnit().contains("percentage")) {
					att.setUnit("%");
				}
				if (availabilityNfr.getUnit().contains("milliseconds")) {
					att.setUnit("ms");
				}
				attributes.add(att);
			}
		}
		if (ms.getClassification().contentEquals("Computing Public IP")) {
		Attribute att = new Attribute();
		att.setTypeid(9);
		att.setName("Public IP");
		att.setValue("IP");
		att.setUnit(null);
		attributes.add(att);
		}
		//pending eliminate nameVM
		//String nameVM = "nononono";
		filterexample.setAttributes(attributes);

		//ObjectWriter ow = new ObjectMapper().writer().withDefaultPrettyPrinter();
        String filter= "";
		//filter = String.valueOf(filterexample);
		
        ObjectMapper mapper = new ObjectMapper();
        mapper.configure(SerializationFeature.INDENT_OUTPUT, true);
        try {
			filter = mapper.writeValueAsString(filterexample);
		} catch (JsonProcessingException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
        
//        try {
//			filter = ow.writeValueAsString(filterexample);
//
//		} catch (JsonProcessingException e) {
//			e.printStackTrace();
//		}
		List<Service> response = RequestToACSmI(filter);
        if (!response.isEmpty()) {
        	for (Service cs : response) {
        		msANDcs mscs = new msANDcs();
        		mscs.setMsid(ms.getId());
        		mscs.setCsid(cs.getId().toString());
        		mscs.setWeight(cs.getMatch());
        		mscs.setMatchingList(cs.getMatchingattributelist());
        	// detachable_resources are always associated to the microservice and they will be deployed in the same cs
        		List<String> detach = new ArrayList<>();
        		for (DetachableResource dr : ms.getDetachableResources()) {
        		detach.add(dr.getId());
        		}
        		mscs.setdrid(detach);
            	schemasPossibilities.add(mscs);
         }

	}

	}

	private List<Service> RequestToACSmI(String filter) {
		//System.out.println("request built is: "+filter);
		JacksonJsonProvider provider = new JacksonJsonProvider();
        List providers = new ArrayList();
        providers.add(provider);
        
        //the integration enviroment
        api = JAXRSClientFactory.create("acsmiServicesURL", ServiceResourceApi.class, providers);
        
        org.apache.cxf.jaxrs.client.Client client = WebClient.client(api);
        
        ClientConfiguration config = WebClient.getConfig(client); 
        List<Service> response = api.findServicesUsingGET(filter);
        return response;
	}
	private void wait (int segundos) {
		try {
			Thread.sleep(segundos * 1000);
		} catch (InterruptedException ex) {
			Thread.currentThread().interrupt();
		}
	}
    private Properties loadProperties() {
        Properties props;
        props = new Properties();
        
        InputStream fis = null;
        try {
            fis = this.getClass().getClassLoader().getResourceAsStream("/optimus.server.properties");
            props.load(fis);
        }     catch (IOException e) {
            e.printStackTrace();
        }
        return props;
  }
}

